{# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0 #}
{# For details: https://bitbucket.org/ned/coveragepy/src/default/NOTICE.txt #}

<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    {# IE8 rounds line-height incorrectly, and adding this emulateIE7 line makes it right! #}
    {# http://social.msdn.microsoft.com/Forums/en-US/iewebdevelopment/thread/7684445e-f080-4d8f-8529-132763348e21 #}
    <meta http-equiv="X-UA-Compatible" content="IE=emulateIE7" />
    <title>Coverage for {{fr.relative_filename|escape}}: {{nums.pc_covered_str}}%</title>
    <link rel="stylesheet" href="style.css" type="text/css">
    {% if extra_css %}
        <link rel="stylesheet" href="{{ extra_css }}" type="text/css">
    {% endif %}
    <script type="text/javascript" src="jquery.min.js"></script>
    <script type="text/javascript" src="jquery.hotkeys.js"></script>
    <script type="text/javascript" src="jquery.isonscreen.js"></script>
    <script type="text/javascript" src="coverage_html.js"></script>
    <script type="text/javascript">
        jQuery(document).ready(coverage.pyfile_ready);
    </script>
</head>
<body class="pyfile">

<div id="header">
    <div class="content">
        <h1>Coverage for <b>{{fr.relative_filename|escape}}</b> :
            <span class="pc_cov">{{nums.pc_covered_str}}%</span>
        </h1>

        <img id="keyboard_icon" src="keybd_closed.png" alt="Show keyboard shortcuts" />

        <h2 class="stats">
            {{nums.n_statements}} statements &nbsp;
            <span class="{{c_run}} shortkey_r button_toggle_run">{{nums.n_executed}} run</span>
            <span class="{{c_mis}} shortkey_m button_toggle_mis">{{nums.n_missing}} missing</span>
            <span class="{{c_exc}} shortkey_x button_toggle_exc">{{nums.n_excluded}} excluded</span>

            {% if has_arcs %}
                <span class="{{c_par}} shortkey_p button_toggle_par">{{nums.n_partial_branches}} partial</span>
            {% endif %}
        </h2>
    </div>
</div>

<div class="help_panel">
    <img id="panel_icon" src="keybd_open.png" alt="Hide keyboard shortcuts" />
    <p class="legend">Hot-keys on this page</p>
    <div>
    <p class="keyhelp">
        <span class="key">r</span>
        <span class="key">m</span>
        <span class="key">x</span>
        <span class="key">p</span> &nbsp; toggle line displays
    </p>
    <p class="keyhelp">
        <span class="key">j</span>
        <span class="key">k</span> &nbsp; next/prev highlighted chunk
    </p>
    <p class="keyhelp">
        <span class="key">0</span> &nbsp; (zero) top of page
    </p>
    <p class="keyhelp">
        <span class="key">1</span> &nbsp; (one) first highlighted chunk
    </p>
    </div>
</div>

<div id="source">
    <table>
        <tr>
            <td class="linenos">
{% for line in lines -%}
    <p id="n{{line.number}}" class="{{line.class}}"><a href="#n{{line.number}}">{{line.number}}</a></p>
{% endfor %}
            </td>
            <td class="text">
{# These are the source lines, which are very sensitive to whitespace. -#}
{# The `{ # - # }` below are comments which slurp up the following space. -#}
{% for line in lines -%}
    <p id="t{{line.number}}" class="{{line.class}}">{#-#}
        {% if line.annotate -%}
            <span class="annotate short">{{line.annotate}}</span>{#-#}
            <span class="annotate long">{{line.annotate_long}}</span>{#-#}
        {% endif -%}
        {{line.html}}<span class="strut">&nbsp;</span>{#-#}
    </p>
{% endfor %}
            </td>
        </tr>
    </table>
</div>

<div id="footer">
    <div class="content">
        <p>
            <a class="nav" href="index.html">&#xab; index</a> &nbsp; &nbsp; <a class="nav" href="{{__url__}}">coverage.py v{{__version__}}</a>,
            created at {{ time_stamp }}
        </p>
    </div>
</div>

</body>
</html>
